// This file is part of the AspectC++ compiler 'ac++'.
// Copyright (C) 1999-2003  The 'ac++' developers (see aspectc.org)
//
// This program is free software;  you can redistribute it and/or
// modify it under the terms of the GNU General Public License as
// published by the Free Software Foundation; either version 2 of
// the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public
// License along with this program; if not, write to the Free
// Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
// MA  02111-1307  USA

#include "CodeWeaver.h"
#include "AspectInfo.h"
#include "AdviceInfo.h"
#ifdef ACMODEL
#include "Utils.h"
#else
#include "JoinPointLoc.h"
#endif
#include "JoinPointPlan.h"
#include "JPAdvice.h"
#include "ACUnit.h"
#include "Naming.h"
#include "IntroductionUnit.h"
#include "NamespaceAC.h"
#include "ResultBuffer.h"

#include "Puma/ErrorSink.h"
#include "Puma/ACTree.h"
#include "Puma/ACAspectInfo.h"
#include "Puma/CProtection.h"
#include "Puma/CTemplateInfo.h"
#include "Puma/CClassInfo.h"
#include "Puma/CUnionInfo.h"
#include "Puma/CAttributeInfo.h"
#include "Puma/CFileInfo.h"
#include "Puma/CArgumentInfo.h"
#include "Puma/CFunctionInfo.h"
#include "Puma/CProject.h"
#include "Puma/FileUnit.h"
#include "Puma/CSemDatabase.h"

void CodeWeaver::type_check (CRecord *rec, const string &name, bool result) {

  // are we in a class (and not in a struct)?
  bool in_class = (rec->ClassInfo () && !rec->ClassInfo ()->isStruct ());

  // create the code
  ostringstream code;
  code << endl;
  if (in_class)
    code << "public:" << endl;
  code << "  virtual bool " << name << " () const { return "
       << (result ? "true" : "false") << "; }"	<< endl;
  if (in_class)
    code << "private:" << endl;

  // paste the function after "{" of the class/struct/union definition
  Token *inspos = ((CT_ClassDef*)rec->Tree ())->Members ()->token ();
  const WeavePos &pos = weave_pos (inspos, WeavePos::WP_AFTER);
  paste (pos, code.str ());
}

#ifdef ACMODEL
void CodeWeaver::add_aspect_include (ACM_Any *jpl, AspectInfo &aspect_info,
#else
void CodeWeaver::add_aspect_include (JoinPointLoc *jpl, AspectInfo &aspect_info,
#endif
  AspectRef::Kind kind) {
  Unit *iu;

#ifdef ACMODEL
  if (jpl && is_pseudo(*jpl))
    return;

  ACM_Aspect &jpl_aspect = aspect_info.loc ();
#else
  if (jpl && jpl->type () == JoinPointLoc::MethodCall &&
      ((JPL_Code*)jpl)->is_pseudo ())
    return;

  JPL_Aspect &jpl_aspect = aspect_info.loc ();
#endif
  // TODO: find out what this 'if' is good for ;-)
  if (!jpl)
    // TODO: isn't there a _primary?
    iu = TI_Aspect::of (jpl_aspect)->ClassInfo ()->SourceInfo ()->FileInfo ()->Primary ();
  else {
    iu = TransformInfo::unit (*jpl);
  }

  // if the 'insert unit' is an 'introduction unit', find the intro target unit
  IntroductionUnit *intro_unit = 0;
  while ((intro_unit = IntroductionUnit::cast (iu)) != 0)
    iu = intro_unit->target_unit ();

  _aspect_includes.insert (iu, &aspect_info, kind);
}

void CodeWeaver::aspect_includes (CProject &project) {

  // weave the aspect includes in all registered unit
  for (AspectIncludes::const_iterator iter = _aspect_includes.begin ();
    iter != _aspect_includes.end (); ++iter) {

    // determine the insertion position for the current unit
    Unit *unit = _aspect_includes.unit (iter);

    // generate the includes for the current unit
    string includes = _aspect_includes.generate (iter, problems ());

    // paste the includes (and optional guard head) at the beginning of the unit
    paste_first (unit, includes);
  }
}

void CodeWeaver::slice_includes (CProject &project, Token *first) {
  Unit *primary = (Unit*)first->belonging_to ();

  // weave the aspect includes in all registered unit
  for (SliceIncludes::const_iterator iter = _slice_includes.begin ();
    iter != _slice_includes.end (); ++iter) {

    // avoid a manipulation error, because the directive is unbalanced
    ignore_unbalanced ();

    // determine the insertion position for the current unit
    Unit *unit = _slice_includes.unit (iter);
    bool is_primary = (unit == primary);
    Token *inspos = is_primary ? first : (Token*)unit->first ();

    // Find out if this unit has already been handled in the first phase.
    // In this case no include guard has to be generated.
    bool guard = (!is_primary);

    // generate the includes for the current unit
    string includes = _slice_includes.generate (project, iter);

    // generate the include guard prologue for the header file
    if (guard) {
      ostringstream guard_head;
      guard_head << endl << "#ifndef ";
      Naming::guard (guard_head, (FileUnit*)unit);
      guard_head << endl << "#define ";
      Naming::guard (guard_head, (FileUnit*)unit);
      guard_head << endl;
      includes += guard_head.str ();
    }

    // paste the includes (and optional guard head) at the beginning
    paste (weave_pos (inspos, WeavePos::WP_BEFORE), includes);

    // generate the #endif for the guard if needed
    if (guard) {
      ostringstream guard_tail;
      guard_tail << endl << "#endif // ";
      Naming::guard (guard_tail, (FileUnit*)unit);
      guard_tail << endl;
      paste (weave_pos ((Token*)unit->last (), WeavePos::WP_AFTER),
        guard_tail.str ());
    }

    // re-enable the manipulator preprocessor directive balance check
    check_unbalanced ();
  }
}

void CodeWeaver::insert_namespace_ac_before (Token *inspos) {
  paste (weave_pos (inspos, WeavePos::WP_BEFORE),
    NamespaceAC::def (_problems._size_type));
}


void CodeWeaver::to_class (ACAspectInfo *ai) {
  const WeavePos &from =
    weave_pos (ai->ClassInfo ()->Tree ()->token (), WeavePos::WP_BEFORE);
  const WeavePos &to =
    weave_pos (ai->ClassInfo ()->Tree ()->token (), WeavePos::WP_AFTER);
  replace (from, to, "class");
}


void CodeWeaver::invocation_functions (ACAspectInfo *ai,
                                       const string &decls,
                                       const string &defs) {

//  Unit *aspect_unit  = (Unit*)ai->ClassInfo ()->SourceInfo ()->StartToken ()->
//                       token ()->belonging_to ();
//  Unit *primary_unit = (Unit*)first->belonging_to ();
//
//  const WeavePos &decl_pos = weave_pos ((aspect_unit == primary_unit) ?
//                                         first : (Token*)aspect_unit->first (),
//                                         WeavePos::WP_BEFORE);
//  paste (decl_pos, decls);

  const WeavePos &def_pos = weave_pos (((CT_ClassDef*)ai->ClassInfo ()->
                                        Tree ())->ObjDecl ()->end_token (),
                                        WeavePos::WP_AFTER);
  paste (def_pos, defs);
}


void CodeWeaver::declare_function (CFunctionInfo *advice_func,
				   CTree *advice_decl) {
  // Create a member function of the aspect that contains the advice code

  CT_FctDef *fctdef = (CT_FctDef*)advice_func->Tree ();
  CT_ArgDeclList *args = ((CT_FctDeclarator*)fctdef->Declarator ())->
    Arguments ();

  const WeavePos &after_args =
    weave_pos (args->end_token (), WeavePos::WP_AFTER);

  // duplicate the arguments
  copy (args, after_args);

  // create and insert the function name
  ostringstream func_decl;
  func_decl << " {}" << endl << "public: "
            << "void __" << (advice_func->Name () + 1) << " ";
  paste (after_args, func_decl.str ());

  // set the protection
  if (advice_func->Protection () != CProtection::PROT_PUBLIC) {
    paste (weave_pos (fctdef->end_token (), WeavePos::WP_AFTER),
      protection_string (advice_func->Protection ()));
  }
}


void CodeWeaver::declare_friends (CClassInfo *cls, list<CClassInfo*> &friends) {
  CClassInfo *def = (CClassInfo*)cls->DefObject ();

  // check if the class is generated by a macro
  if (def->Tree ()->token ()->is_macro_generated ()) {
    if (_problems._warn_macro) {
      _err << sev_warning << def->Tree ()->token ()->location ()
           << "class '" << def->QualName () << "' defined by macro"
           << endMessage;
    }
    return;
  }

  ostringstream friend_decls;
  ostringstream forward_decls;

  // close namespaces for the forward declarations
  close_namespace (forward_decls, cls);

  // add all forward and friend declarations
  for (list<CClassInfo*>::iterator iter = friends.begin ();
       iter != friends.end (); ++iter) {

    // if this is the aspect itself, skip this one
    if (def == (*iter)->DefObject ())
      continue;

    friend_decls << "  friend class ::" << (*iter)->QualName () << ";" << endl;
    forward_decls << "class " << (*iter)->Name () << ";" << endl;
  }

  // re-open the namespace
  open_namespace (forward_decls, cls);

  // insert the friend declarations into the class
  Token *inspos = ((CT_ClassDef*)def->Tree ())->Members ()->token ();
  paste (weave_pos (inspos, WeavePos::WP_AFTER), friend_decls.str ());

  // generate forward declarations iff this is no nested class
  if (!(nscope (def) && nscope (def)->ClassInfo ())) {
    // insert a forward declaration in front of the class
    inspos = def->TemplateInfo () ?
             def->TemplateInfo ()->Tree ()->token () :
             ((CT_ClassDef*)def->Tree ())->ObjDecl ()->token ();
    paste (weave_pos (inspos, WeavePos::WP_BEFORE), forward_decls.str ());
  }
}


void CodeWeaver::open_namespace (ostream &out, CObjectInfo *obj) {
  CClassInfo *cls = obj->ClassInfo();
  if (cls && cls->TemplateInfo ())
    obj = (CObjectInfo*)cls->TemplateInfo ();

  Array<CObjectInfo*> namespaces;
  while (!obj->Scope ()->GlobalScope () && obj->Scope ()->NamespaceInfo ()) {
    obj = obj->Scope ();
    namespaces.append (obj);
  }

  for (int i = namespaces.length () - 1; i >= 0; i--) {
    if (namespaces[i]->isAnonymous ())
      out << "namespace {" << endl;
    else
      out << "namespace " << namespaces[i]->Name () << " {" << endl;
  }
}


void CodeWeaver::close_namespace (ostream &out, CObjectInfo *obj) {
  CClassInfo *cls = obj->ClassInfo();
  if (cls && cls->TemplateInfo ())
    obj = (CObjectInfo*)cls->TemplateInfo ();

  ostringstream ns_close;
  while (!obj->Scope ()->GlobalScope () && obj->Scope ()->NamespaceInfo ()) {
    obj = obj->Scope ();
    out << "} // closed " << (obj->isAnonymous () ? "" : obj->QualName ()) << endl;
  }
}


void CodeWeaver::singleton (ACAspectInfo *ai) {
  if (ai->aspect_of ()) {
    // if necessary, rename aspectOf to aspectof
    if (strcmp (ai->aspect_of ()->Name (), "aspectOf") == 0) {
      CT_FctDef *fctdef = (CT_FctDef*)ai->aspect_of ()->Tree ();
      CT_SimpleName *name = ((CT_Declarator*)fctdef->Declarator ())->Name ();
      name = name->Name (); // if it is qualified
      const WeavePos &name_start =
        weave_pos (name->token (), WeavePos::WP_BEFORE);
      const WeavePos &name_end =
        weave_pos (name->end_token (), WeavePos::WP_AFTER);
      replace (name_start, name_end, "aspectof");
    }
  }
  else {
    // paste __instance declaration after "{" of the aspect definition
    Token *member_list = ((CT_ClassDef*)ai->ClassInfo ()->Tree ())->
      Members ()->token ();
    const WeavePos &pos = weave_pos (member_list, WeavePos::WP_AFTER);

    ostringstream func;
    func << endl
      << "public:" << endl
      << "  static " << ai->name () << " *aspectof () {" << endl
      << "    static " << ai->name () << " __instance;" << endl
      << "    return &__instance;" << endl
      << "  }" << endl
      << "  static " << ai->name () << " *aspectOf () {" << endl
      << "    return aspectof ();" << endl
      << "  }" << endl
      << "private:" << endl;

    paste (pos, func.str ());
  }
}

// generate the code, which proceeds an intercepted flow of control, e.g.
// continues a call or execution
#ifdef ACMODEL
void CodeWeaver::make_proceed_code (ostream &out, ACM_Any *badly_typed_loc,
                                    bool action, int max_depth) {
  ACM_Code *loc = (ACM_Code*)badly_typed_loc;
#else
void CodeWeaver::make_proceed_code (ostream &out, JoinPointLoc *badly_typed_loc,
                                    bool action, int max_depth) {
  JPL_Code *loc = (JPL_Code*)badly_typed_loc;
#endif
  TI_Code *ti = (TI_Code*)TransformInfo::of (*loc);

  // set flags that describe the kind of join point
  bool is_exec = false, is_call = false, is_cons = false, is_dest = false;
#if ACMODEL
  if (loc->type_val () == JPT_Execution)
    is_exec = true;
  else if (loc->type_val () == JPT_Call)
    is_call = true;
  else if (loc->type_val () == JPT_Construction)
    is_cons = true;
  else if (loc->type_val () == JPT_Destruction)
    is_dest = true;
  else
    return; // should not happen!

  // find out which function has to be called by proceed
  ACM_Function *func = 0;
  if (is_call)
    func = ((ACM_Call*)loc)->get_target();
  else if (is_exec || is_cons || is_dest)
    func = (ACM_Function*)loc->get_parent ();
#else
  if (loc->type () == JoinPointLoc::Method)
    is_exec = true;
  else if (loc->type () == JoinPointLoc::MethodCall)
    is_call = true;
  else if (loc->type () == JoinPointLoc::Construction)
    is_cons = true;
  else if (loc->type () == JoinPointLoc::Destruction)
    is_dest = true;
  else
    return; // should not happen!

  // find out which function has to be called by proceed
  JPL_Function *func = 0;
  if (is_call)
    func = ((JPL_MethodCall*)loc)->target_function ();
  else if (is_exec || is_cons || is_dest)
    func = (JPL_Function*)loc->parent ();
#endif

  // analyse the target object of this call
  bool target_is_ptr = false; // initialization to avoid warning
  CT_Expression *target_expr = 0;
  if (is_call)
    target_expr = ((TI_MethodCall*)ti)->target_expr (target_is_ptr);

  CFunctionInfo *dstfunc = ((TI_Function*)func->transform_info())->func_info ();
  bool obj_needed = needs_this (dstfunc) || (target_expr && !action);

  // nothing is done to proceed implicitly defined special member fcts
  if ((is_cons || is_dest) && dstfunc->isBuiltin ())
    return;

  // create a local class that enters/leaves the cflow triggers
  const CFlowList &cflows = _code_plan->cflows ();
  if (cflows.size () > 0) {
    out << "{" << endl;
    cflows.gen_trigger_obj (out);
  }

  // generate the expression that calculates the result
  stringstream res;

  bool bypass = (is_call && ((TI_MethodCall*)ti)->needs_rights());
  CClassInfo *dstcls = cscope (dstfunc);
  if (bypass) {
    // check priviledged calls to external functions
    Unit *unit = dstcls->SourceInfo ()->SrcUnit ();
    bool dst_external = false;
    if (!dstcls->SemDB ()->Project ()->isBelow (unit) &&
        !IntroductionUnit::cast (unit)) {
      dst_external = true;
    }

    if (dst_external) {
      // calls to private externals are not supported
      // target function is protected
      if (target_expr) {
        CTypeInfo *target_type = target_expr->Type ();
        while (target_type->TypeAddress ())
          target_type = target_type->BaseType ();
        dstcls = target_type->ClassInfo();
      }
      else {
        dstcls = cscope (ti->assoc_obj());
      }
    }

    res << "::" << dstcls->QualName() << "::Caller<";
#ifdef ACMODEL
    Naming::bypass_id_class (res, (ACM_Call*)loc);
#else
    Naming::bypass_id_class (res, (JPL_MethodCall*)loc);
#endif
    res << ">::call(";
  }

  int args = 0;
  // if an object is needed for the call, issue 'obj->'
  if (obj_needed) {
    if (is_call) {
      if (bypass) {
        res << "(::" << dstcls->QualName() << "*)";
        args++;
      }
      if (action)
        res << "__TJP::target ()";
      else {
        res << "dstthis";
      }
      if (!bypass)
        res << "->";
    }
    else if (is_exec || is_cons || is_dest) {
      if (action)
        res << "__TJP::that ()->";
      else
        res << "this->";
    }
  }

  // generate the function name
  if (is_exec || (!dstfunc->isBuiltin () && (is_cons || is_dest))) {
    if (!obj_needed && nscope (dstfunc)) {
      string scope_name = string ("::") + nscope (dstfunc)->QualName ();
      res << cleanup_name(scope_name) << "::";
    }
    // use __old_<name> for execution join points
    Naming::exec_inner (res, loc);
  } else if (is_call && !bypass) {
    // destination function name
    if (((TI_MethodCall*)TransformInfo::of (*loc))->is_qualified () ||
        dstfunc->isStaticMethod () || !dstfunc->isMethod ()) {
      string name = cleanup_name (dstfunc->QualName (true, true));
      res << name;
    }
    else
      res << dstfunc->Name ();

    // add the template arguments if it is a function template instance
    CTemplateInstance *instance = dstfunc->TemplateInstance ();
    if (instance) {
      res << "< ";
      for (unsigned a = 0; a < instance->DeducedArgs (); a++) {
        if (a > 0) res << ",";
        DeducedArgument *arg = instance->DeducedArg (a);
        if (arg->Type ())
          arg->Type ()->TypeText (res, "", true, true);
        else if (arg->Value ()) {
          if (arg->Value ()->isSigned ())
            res << arg->Value ()->convert_to_int ();
          else if (arg->Value ()->isUnsigned ())
            res << arg->Value ()->convert_to_uint ();
          else if (arg->Value ()->isFloat ())
            res << arg->Value ()->convert_to_float ();
        }
        else
          res << "*invalid template arg*";
      }
      res << " >";
    }
  }

  // destination function arguments
  if (!bypass)
    res << "(";
#ifdef ACMODEL
  for (int a = 0 ; a < get_arg_count (*loc); a++, args++) {
#else
  for (int a = 0 ; a < loc->arg_count (); a++, args++) {
#endif
    if (args > 0) res << ", ";
    if (action) {
      // make sure that if the argument type was a reference we use the
      // referred type here
      res << "*(typename __TJP::template Arg<" << a
          << ">::ReferredType*)__TJP::arg(" << a << ")";
    }
    else
      res << "arg" << a;
  }
  res << ")";

  // generate the 'result =' code
  CTypeInfo *result_type =
    dstfunc->isConversion () ? dstfunc->ConversionType () :
    dstfunc->TypeInfo ()->BaseType ();

  ResultBuffer result_buffer (result_type);
  out << "    ";
  if (action)
    out << result_buffer.action_result_assignment(res.str ());
  else
    out << result_buffer.result_assignment(res.str ());

  out << ";" << endl;

  if (cflows.size () > 0)
    out << "}" << endl;
}

// create an action wrapper function, which may be invoked by 'proceed()'
#ifdef ACMODEL
void CodeWeaver::make_action_wrapper(ostream &impl, ACM_Any *loc,
#else
void CodeWeaver::make_action_wrapper(ostream &impl, JoinPointLoc *loc,
#endif
    JPAdvice *jpa) {
  // generate the action function, which is used to obtain an action object
  impl << "  AC::Action &action() {" << endl;
  impl << "    this->__TJP::_wrapper = &";
  Naming::action_wrapper (impl, loc, jpa->depth ());
  impl << ";" << endl;
  impl << "    return *this;" << endl;
  impl << "  }" << endl;

  // make the function static if it is defined in class scope
  impl << "  static void ";
  Naming::action_wrapper (impl, loc, jpa->depth ());
  impl << " (AC::Action &action) {" << endl;
  // convert the action object ref into a TJP_... object ref and call proceed
  impl << "    ((__TJP&)action).proceed ();" << endl;

  impl << "  }" << endl;
}

// create the 'proceed()' function for the current level
#ifdef ACMODEL
void CodeWeaver::make_proceed_func(ostream &impl, ACM_Any *loc,
#else
void CodeWeaver::make_proceed_func(ostream &impl, JoinPointLoc *loc,
#endif
    JPAdvice *jpa) {
  // the signature
  impl << "  void proceed () {" << endl;

  if (jpa->depth() == 0) {
    // this is a proceed function for the last around advice
    make_proceed_code (impl, loc, true, jpa->max_depth ());
  }
  else {
    // generate __TJP type (for the next level)
    impl << "    typedef ";
#ifdef ACMODEL
    Naming::tjp_struct (impl, (ACM_Code*)loc, jpa->depth () - 1);
#else
    Naming::tjp_struct (impl, (JPL_Code*)loc, jpa->depth () - 1);
#endif
    impl << "<TResult, TThat, TTarget, TArgs> __TJP;" << endl;
    // generate advice calls
    make_advice_calls (impl, jpa->next (), loc, true);
  }
  impl << "  }" << endl;
}

// generate the name of the joinpoint-specific TJP type
#ifdef ACMODEL
void CodeWeaver::make_tjp_typename (ostream &out, ACM_Code *loc, int depth) {
#else
void CodeWeaver::make_tjp_typename (ostream &out, JPL_Code *loc, int depth) {
#endif
  TI_Code *ti = (TI_Code*)TransformInfo::of (*loc);
  Naming::tjp_struct (out, loc, depth);
  out << "< " << ThisJoinPoint::gen_result_type (loc) << ", ";
  ti->that_type ()->TypeText (out, "", true, true);
  out << ", ";
  ti->target_type ()->TypeText (out, "", true, true);
  // argument types
#ifdef ACMODEL
  unsigned arg_count = get_arg_count (*loc);
#else
  unsigned arg_count = loc->arg_count ();
#endif
  out << ", ";
#ifdef ACMODEL
  list<ACM_Type*> arg_types;
  get_arg_types (*loc, arg_types);
  for (list<ACM_Type*>::iterator i = arg_types.begin(); i != arg_types.end(); ++i) {
    CTypeInfo *argtype = TI_Type::of (**i)->type_info ();
#else
  for (unsigned a = 0; a < arg_count; a++) {
    CTypeInfo *argtype = TI_Type::of (loc->arg_type (a))->type_info ();
#endif
    out << " AC::TL< ";
    argtype->TypeText (out, "", true, true);
    out << ",";
  }
  out << " AC::TLE";
  for (unsigned a = 0; a < arg_count; a++)
    out << " >";
  out << " >";
 }

// insert TJP class definition before join point
#ifdef ACMODEL
void CodeWeaver::make_tjp_struct(ostream &out, ACM_Any *ploc,
                                 JPAdvice *jpa) {
    ACM_Code *loc = (ACM_Code*)ploc;
#else
void CodeWeaver::make_tjp_struct(ostream &out, JoinPointLoc *ploc,
                                 JPAdvice *jpa) {
    JPL_Code *loc = (JPL_Code*)ploc;
#endif

  // recursively call the function for the previous TJP classes
  if (jpa->depth () > 0) {
    make_tjp_struct (out, loc, jpa->next ());
  }

  // generate the definition
  out << endl;
  const ThisJoinPoint &tjp = jpa->root ()->tjp ();
  tjp.gen_tjp_struct (out, loc, _problems, jpa->depth ());

  // generate the proceed function for the current structure
  if (jpa->around ()) {
    const ThisJoinPoint &curr_tjp =
      TI_AdviceCode::of (jpa->around ()->code ())->this_join_point ();
    if (curr_tjp.proceed () || curr_tjp.action() ||
        !jpa->around ()->pointcut ().cflow_triggers ().empty ())
      make_proceed_func (out, loc, jpa);

    // generate action wrapper if needed
    if (curr_tjp.action())
      make_action_wrapper (out, loc, jpa);
  }

  if (jpa->depth () == 0) {
    stringstream jpname;
    Naming::tjp_struct(jpname, loc, 0);
    jpa->root ()->gen_binding_templates (out, jpname.str ().data (), _problems);
  }

  // closing bracket of the TJP class! Opened in gen_tjp_struct.
  out << "};" << endl;

  out << endl;
}

#ifdef ACMODEL
void CodeWeaver::make_tjp_common_init(ostream &code, ACM_Any *badly_typed_loc,
              JPAdvice *jpa) {
  jpa->tjp ().gen_tjp_init(code, (ACM_Code*)badly_typed_loc,
    _problems, jpa->depth ());
}
#else
void CodeWeaver::make_tjp_common_init(ostream &code, JoinPointLoc *badly_typed_loc,
              JPAdvice *jpa) {
  jpa->tjp ().gen_tjp_init(code, (JPL_Code*)badly_typed_loc,
    _problems, jpa->depth ());
}
#endif

// generates the signature of a wrapper function for exec/cons/dest join pts
#ifdef ACMODEL
string CodeWeaver::wrapper_function_signature (ACM_Code *loc,
#else
string CodeWeaver::wrapper_function_signature (JPL_Code *loc,
#endif
                                               CFunctionInfo *func, bool def) {
  assert (func);

  stringstream name_args;
  if (func->CObjectInfo::QualifiedScope () &&
      func->Scope () != func->CObjectInfo::QualifiedScope ())
    name_args << func->CObjectInfo::QualifiedScope ()->QualName () << "::";
  Naming::exec_inner (name_args, loc);
  name_args << "(";
  for (unsigned a = 0; a < func->Arguments (); a++) {
    CArgumentInfo *arg = func->Argument (a);
    if (arg->TypeInfo ()->is_void ())
      break;
    if (a > 0) name_args << ",";
      arg->TypeInfo ()->TypeText (name_args,
                                  arg->isAnonymous () ? "" : arg->Name (),
                                  true, true);
  }
  name_args << ")";

  ostringstream wrapped;

  if (_problems._use_always_inline) // GNU extension
    wrapped << "__attribute__((always_inline)) ";
  wrapped << "inline "; // the wrapped function should always be inlined
  if (func->isStaticMethod () && func->ClassScope () == func->Scope ())
    wrapped << "static ";

  if (func->isConstructor () || func->isDestructor ())
    wrapped << "void " << name_args.str ().c_str ();
  else {
    CTypeInfo *type = func->isConversion () ? func->ConversionType () :
                                              func->TypeInfo ()->BaseType ();
      type->TypeText (wrapped, name_args.str ().c_str (), true, true);
    if (func->TypeInfo ()->isConst ())
      wrapped << " const";
    if (func->TypeInfo ()->isVolatile ())
      wrapped << " volatile";
  }

  return wrapped.str ();
}

#ifdef ACMODEL
void CodeWeaver::cons_join_point (ACM_Construction *loc, JPP_Code &plan) {
  if (((ACM_Name*)loc->get_parent ())->get_builtin())
#else
void CodeWeaver::cons_join_point (JPL_Construction *loc, JPP_Code &plan) {
  if (loc->function ()->is_built_in())
#endif
    gen_special_member_function (loc, plan);
  else
    wrap_function (loc, plan);
}

#ifdef ACMODEL
void CodeWeaver::dest_join_point (ACM_Destruction *loc, JPP_Code &plan) {
  if (((ACM_Name*)loc->get_parent ())->get_builtin())
#else
void CodeWeaver::dest_join_point (JPL_Destruction *loc, JPP_Code &plan) {
  if (loc->function ()->is_built_in())
#endif
  gen_special_member_function (loc, plan);
  else
    wrap_function (loc, plan);
}


// checks whether an implicitely defined constructor or destructor may be
// generated
bool CodeWeaver::check_special_member_function (CFunctionInfo *func) {
  // the argument should be a valid class member
  assert (func && func->ClassScope ());
  CClassInfo *cls = func->ClassScope ()->ClassInfo ();

  // check if any base class has a private constructor/copy constructor/destr.
  for (unsigned b = 0; b < cls->BaseClasses (); b++) {
    CClassInfo *base = cls->BaseClass (b)->Class ();
    for (unsigned f = 0; f < base->Functions (); f++) {
      CFunctionInfo *curr = base->Function (f);
      if ((curr->isConstructor () && func->isConstructor () &&
           curr->Arguments () == func->Arguments ()) ||
          (curr->isDestructor () && func->isDestructor ())) {
        if (curr->Protection () == CProtection::PROT_PRIVATE ||
            !check_special_member_function (curr))
          return false; // function or any base is private -> problem!
      }
    }
  }
  return true; // no problem
}


#ifdef ACMODEL
void CodeWeaver::gen_special_member_function (ACM_Any *loc,
#else
void CodeWeaver::gen_special_member_function (JoinPointLoc *loc,
#endif
                                              JPP_Code &plan) {
  _code_plan = &plan;

  ostringstream code;
  JPAdvice *ejpa = plan.jp_advice ();
  CFunctionInfo *func = TransformInfo::of (*loc)->assoc_obj ()->FunctionInfo ();
  assert (func && func->ClassScope ());
  CClassInfo *cls = func->ClassScope ()->ClassInfo ();

  // check if the special member would have to call a private member
  if (!check_special_member_function (func))
    return; // can be ignored, because the implicit special member can't be
            // called anyway

  Token *inspos = ((CT_ClassDef*)cls->Tree ())->Members ()->end_token ();

  // create thisJoinPoint class
  ejpa->mergeTJPFlags();
  if (ejpa->tjp ().type_needed ()) {
    ostringstream wrappers;
    make_tjp_struct(wrappers, loc, ejpa);
    paste (weave_pos (inspos, WeavePos::WP_BEFORE), wrappers.str ());
  }

  // generate the function itself
  code << endl << "public:" << endl << "inline " << func->Name ();

  // constructors or destructors don't have more than one argument
  assert (func->Arguments () <= 1);

  if (func->Arguments () == 0) {
    // default constructor or destructor
    code << " () {" << endl;
  }
  else {
    CT_ClassDef *cd = (CT_ClassDef*)cls->Tree ();
    // wrap attributes that are non-static arrays in a class.
    wrap_attribute_arrays (cd->Members ());
    // the same for introduced members
    if (cd->IntroMembers ()) {
      wrap_attribute_arrays ((CT_MembList*)cd->IntroMembers ());
    }

    // generate the copy constructor
    code << " (";
    func->Argument (0u)->TypeInfo ()->TypeText (code, "arg0");
    code << ") ";
    // copy the baseclass members
    bool first_initializer = true;
    for (unsigned b = 0; b < cls->BaseClasses (); b++) {
      if (first_initializer) {
        first_initializer = false;
        code << ": ";
      }
      else
        code << ", ";
      CClassInfo *base = cls->BaseClass (b)->ClassPseudoInstance () ?
        cls->BaseClass (b)->ClassPseudoInstance () :
        cls->BaseClass (b)->Class ();
      code << *base->TypeInfo () << " (arg0)";
    }

    set<CUnionInfo*> anon_unions;
    for (unsigned a = 0; a < cls->Attributes (); a++) {
      CAttributeInfo *attr = cls->Attribute (a);

      // only certain members are initialized
      if (attr->isStatic () || attr->isAnonymous () || attr->EnumeratorInfo ())
        continue;

      // attributes that are member of anonymous unions have to be copied
      // differently
      if (attr->Scope ()->isUnion ()) {
        anon_unions.insert (attr->Scope ()->UnionInfo ());
        continue;
      }

      // write ':' or ','
      if (first_initializer) {
        first_initializer = false;
        code << ": ";
      }
      else
        code << ", ";
      // initialize the member by copy-constructor
      code << attr->Name () << " (arg0." << attr->Name () << ")";
    }
    code << " {" << endl;

    // copy anonymous unions
    int count = 0;
    for (set<CUnionInfo*>::const_iterator iter = anon_unions.begin ();
      iter != anon_unions.end (); ++iter, ++count) {
      CUnionInfo *curr_union = *iter;
      CAttributeInfo *attr = curr_union->Attribute (0);
      // anonymous unions are very simple to copy. They have no base classes,
      // no function members, no static members, ...

      // create a named local copy of the union definition
      code << "  union __ac_union_" << count << " ";
      print_tree (code, ((CT_UnionDef*)curr_union->Tree ())->Members ());
      code << ";" << endl;
      code << "  *(union __ac_union_" << count << "*)&" << attr->Name ()
           << " = *(union __ac_union_" << count << "*)&arg0."
           << attr->Name () << ";" << endl;
    }
  }

  // generate __TJP type (for this level)
  if (ejpa->tjp ().type_needed ()) {
    code << "  typedef ";
#ifdef ACMODEL
    make_tjp_typename (code, (ACM_Code*)loc, ejpa->depth ());
#else
    make_tjp_typename (code, (JPL_Code*)loc, ejpa->depth ());
#endif
    code << " __TJP;" << endl;
  }

  // generate common JoinPoint initialization
  make_tjp_common_init(code, loc, ejpa);

  // generate calls to advice code or original function
  make_advice_calls(code, ejpa, loc);

  code << endl << "}" << endl;

  // insert the definition
  paste (weave_pos (inspos, WeavePos::WP_BEFORE), code.str ());
}

// wrap attributes that are non-static arrays in a class.
void CodeWeaver::wrap_attribute_arrays (CT_MembList *members) {
  for (int m = 0; m < members->Entries (); m++) {
    if (members->Entry (m)->NodeName () != CT_ObjDecl::NodeId ())
      continue;
    CT_ObjDecl *decl = (CT_ObjDecl*)members->Entry (m);
    CT_DeclSpecSeq *decl_specs = decl->DeclSpecs ();
    CT_DeclaratorList *dlist = decl->Declarators ();
    bool del_decl_specs = false;
    for (int init_decls = 0, s = 0; s < dlist->Sons (); s++) {
      if (dlist->Son (s)->NodeName () != CT_InitDeclarator::NodeId ())
        continue;
      CT_InitDeclarator *init_decl = (CT_InitDeclarator*)dlist->Son (s);
      CObjectInfo *obj = init_decl->SemObject ()->Object ();

      // if this object is a typedef, nothing is of interest here
      if (obj->TypedefInfo ())
        break;

      // replace comma with semi colon
      if (init_decls > 0) {
        // replace the "," of the init declarator list with ";" if needed
        CTree *comma = dlist->Son (s - 1);
        const WeavePos &from = weave_pos (comma->token (), WeavePos::WP_BEFORE);
        const WeavePos &to   = weave_pos (comma->token (), WeavePos::WP_AFTER);
        replace (from, to, ";");
      }

      // only non-static arrays are to be wrapped
      if (obj->TypeInfo ()->TypeArray () && !obj->isStatic ()) {
        // it is a non-static array

        // generate the code of the wrapper class
        ostringstream wrapper_class_code;
        gen_wrapped_array (wrapper_class_code, obj);

        // replace the original init declarator with the wrapper object decl
        const WeavePos &from =
          weave_pos (init_decl->token (), WeavePos::WP_BEFORE);
        const WeavePos &to =
          weave_pos (init_decl->end_token (), WeavePos::WP_AFTER);
        replace (from, to, wrapper_class_code.str ());

        // if this is the first init declarator, remove the decl spec seq
        if (init_decls == 0)
          del_decl_specs = true;
      }
      else {
        // it is no array; copy the declaration specifier list
        if (decl_specs && decl_specs->token () && init_decls > 0) {
          const WeavePos &inspos =
            weave_pos (init_decl->token (), WeavePos::WP_BEFORE);
          copy (decl_specs, inspos);
          paste (inspos, " ");
        }
      }

      // this was an init declarator and not a comma
      init_decls++;
    }

    // delete the decl specs
    if (del_decl_specs && decl_specs && decl_specs->token ())
      kill (decl_specs);
  }
}

void CodeWeaver::gen_wrapped_array (ostream &out, CObjectInfo *obj) {
  assert (obj->TypeInfo ()->TypeArray ());
  CTypeArray *t_array = obj->TypeInfo ()->TypeArray ();
  out << endl;
  out << "  struct ";

  // generate name to satisfy VC++ 2002
  out << "__ac_wrapper_" << obj->Name ();

  out << " {" << endl;
  out << "    typedef ";
  t_array->BaseType ()->TypeText (out, "E", true, true);
  out << "; typedef E A[" << t_array->Dimension () << "]; A _data;" << endl;
  out << "    operator A& () { return _data; }" << endl;
  out << "    operator A& () const { return (A&)*(";
  CTypeArray *t_first_array = t_array;
  while (t_first_array->BaseType ()->TypeArray ())
    t_first_array = t_first_array->BaseType ()->TypeArray ();
  t_first_array->BaseType ()->TypeText (out, "*", true, true);
  out << ")_data; }" << endl;
  out << "    operator const A& () { return _data; }" << endl;
  out << "    operator const A& () const { return _data; }" << endl;
  out << "    operator void* () { return _data; }" << endl;
  out << "    operator void* () const { return (void*)_data; }" << endl;
  out << "    operator const void* () { return _data; }" << endl;
  out << "    operator const void* () const { return _data; }" << endl;
  out << "    template <typename I> E& operator [] (I i) "
      << "{ return _data[i]; } // for VC++ 2003" << endl;
  out << "    template <typename I> const E& operator [] (I i) const "
      << "{ return _data[i]; } // for VC++ 2003" << endl;
  out << "  } " << obj->Name ();
}

#ifdef ACMODEL
void CodeWeaver::exec_join_point (ACM_Execution *loc, JPP_Code &plan) {
#else
void CodeWeaver::exec_join_point (JPL_Method *loc, JPP_Code &plan) {
#endif
  wrap_function (loc, plan);
}

#ifdef ACMODEL
void CodeWeaver::wrap_function (ACM_Code *loc, JPP_Code &plan) {
#else
void CodeWeaver::wrap_function (JPL_Code *loc, JPP_Code &plan) {
#endif
  _code_plan = &plan;

  CFunctionInfo *func = TransformInfo::of (*loc)->assoc_obj ()->DefObject ()->FunctionInfo ();
  assert (func);

  // handle each declaration and the optional definition separately
  // every object info is stored in a cyclic list
  CFunctionInfo *def = 0;
  bool wrapped_decl = false;
  CFunctionInfo *curr = func;
  do {
    if (curr->isFctDef ())
      def = curr;
    else {
      if (wrap_function_decl (loc, curr)) {
        // friend declarations have an assigned scope. They are no real
        // declarations. Another declaration has to be generated.
        if (!curr->AssignedScope ())
          wrapped_decl = true;
      }
    }
    curr = curr->NextObject ()->FunctionInfo ();
  }
  while (curr != func);

  // if a definition has been found, wrap it
  if (def)
    wrap_function_def (loc, def, wrapped_decl);
}

// create a new function implementation and paste it behind the original
#ifdef ACMODEL
void CodeWeaver::wrap_function_def (ACM_Code *loc, CFunctionInfo *func,
#else
void CodeWeaver::wrap_function_def (JPL_Code *loc, CFunctionInfo *func,
#endif
  bool wrapped_decl) {

  CT_FctDef *fctdef = (CT_FctDef*)func->Tree ();
  JPAdvice *ejpa = _code_plan->jp_advice ();
  ostringstream pre_sig, pre_body;

  bool have_pre_sig = false;
  pre_sig << endl;

  // the first is the declaration of the inner function (the original)
  // not necessary for methods and if there is a declaration anyway
  if (!func->isMethod () && !wrapped_decl) {
    have_pre_sig = true;
    pre_sig << wrapper_function_signature (loc, func, true) << ";" << endl;
  }

  // generate the JoinPoint class
  ejpa->mergeTJPFlags();
  if (ejpa->tjp ().type_needed ()) {
    have_pre_sig = true;
    if (fctdef->Linkage () && fctdef->Linkage ()->isList ())
      pre_sig << "}" << endl << endl;
    make_tjp_struct(pre_sig, loc, ejpa);
    if (fctdef->Linkage () && fctdef->Linkage ()->isList ())
      pre_sig << endl << "extern \"C\" {" << endl;
  }

  // paste pre-signature code in front of the old function signature
  if (have_pre_sig) { // check only for better readability
    Token *pre_sig_token = linkage_adjust (fctdef)->token ();
    paste (weave_pos (pre_sig_token, WeavePos::WP_BEFORE), pre_sig.str ());
  }

  // rename the arguments of the original declaration
  rename_args (func, "arg");

  // generate the body of the wrapper function
  CTypeFunction *ftype = func->TypeInfo ()->TypeFunction ();
  CTypeInfo *result_type = func->isConversion () ?
    func->ConversionType () : ftype->BaseType ();
  ResultBuffer result_buffer (result_type);

  pre_body << "{" << endl;

  // generate __TJP type (for this level)
  if (ejpa->tjp ().type_needed ()) {
    pre_body << "  typedef ";
    make_tjp_typename (pre_body, loc, ejpa->depth ());
    pre_body << " __TJP;" << endl;
  }

  // declare the result object
  pre_body << "  " << result_buffer.result_declaration ();

  // generate common JoinPoint initialization
  make_tjp_common_init (pre_body, loc, ejpa);

  // generate calls to advice code or original function
  make_advice_calls (pre_body, ejpa, loc);

  pre_body << "  " << result_buffer.result_return ();
  pre_body << endl << "}" << endl;

  // generate the signature of the inner function (the original)
  // special treatment for operator new/delete/new[]/delete[]
  if (func->isMethod () && !func->isStaticMethod () && !needs_this (func))
    pre_body << "static ";
  // add a wrapper function declaration
  pre_body << wrapper_function_signature (loc, func, true);

  // insert the new body and new signature in front of the old body
  paste (weave_pos (fctdef->Body ()->token (), WeavePos::WP_BEFORE),
         pre_body.str ());
}

// create a new declaration for a wrapped function
#ifdef ACMODEL
bool CodeWeaver::wrap_function_decl (ACM_Code *loc, CFunctionInfo *func) {
#else
bool CodeWeaver::wrap_function_decl (JPL_Code *loc, CFunctionInfo *func) {
#endif

  // first check if this declaration is built-in
  if (func->isBuiltin ())
    return false;

  // check if the function declaration belongs to the project
  // if it does not, we cannot insert the wrapper declaration
  Unit *unit = func->SourceInfo ()->SrcUnit ();
  if (!func->SemDB ()->Project ()->isBelow (unit) &&
      !IntroductionUnit::cast (unit))
    return false;

  string inner_decl;
  if (func->ClassScope ())
    inner_decl += "public: ";
  // special treatment for friend declarations
  if (func->AssignedScope ())
    inner_decl += "friend ";
  // special treatment for operator new/delete/new[]/delete[]
  if (func->isMethod () && !func->isStaticMethod () && !needs_this (func))
    inner_decl += "static ";
  // add the wrapper function declaration
  inner_decl += wrapper_function_signature (loc, func->FunctionInfo (), false);
  inner_decl += ";";
  // switch to the right protection if needed
  if (func->Protection () != CProtection::PROT_NONE &&
      func->Protection () != CProtection::PROT_PUBLIC)
    inner_decl += protection_string (func->Protection ());
  else
    inner_decl += "\n";

  // move the generated declaration in front of the original declaration
  CT_ObjDecl *objdecl = ((CT_InitDeclarator*)func->Tree ())->ObjDecl ();

  // handle extern "C" declarations
  CTree *pos = linkage_adjust (objdecl);
  paste (weave_pos (pos->token (), WeavePos::WP_BEFORE), inner_decl);

  return true;
}


void CodeWeaver::provide_typename (CTree *node) {
  // check for qualified names
  if (node->NodeName () == CT_SimpleName::NodeId ()) {
    CObjectInfo *obj = ((CT_SimpleName*)node)->Object ();
    if (obj && obj->ClassScope () &&
	      strcmp (obj->ClassScope ()->Name (), "JoinPoint") == 0 &&
	      obj->FunctionInfo () && obj->FunctionInfo ()->Arguments () == 0 &&
	      strcmp (obj->Name (), "arg") == 0) {

      paste (weave_pos (node->token (), WeavePos::WP_BEFORE), "template ");
    }
  }
  else if (node->NodeName () == CT_RootQualName::NodeId () ||
           node->NodeName () == CT_QualName::NodeId ()) {
    // this check is a workaround as long as templates are not instantiated
    CT_QualName *qn = (CT_QualName*)node;
    CT_SimpleName *scp_name = (CT_SimpleName*)qn->Entry (0);
    if (strcmp (scp_name->Text (), "JoinPoint") != 0) {
      // if this qualified name does not start with "JoinPoint" it is not
      // relevant. However, if parts of the qualified name are template names,
      // the template arguments have to be checked and maybe transformed
      for (int e = 0; e < qn->Entries(); e++)
        if (qn->Entry(e)->NodeName() == CT_TemplateName::NodeId())
          provide_typename (((CT_TemplateName*)qn->Entry(e))->Arguments ());
      return;
    }
    // workaround as long as JoinPoint::Arg<N>::Type is not resolved
    if (strcmp (((CT_QualName*)node)->Text (), "Type") == 0 ||
        strcmp (((CT_QualName*)node)->Text (), "ReferredType") == 0) {

      paste (weave_pos (node->token (), WeavePos::WP_BEFORE), "typename ");
    }
    else {
      // replace [::]JoinPoint::(That|Target|Result) with the generated typedef
      if (((CT_QualName*)node)->Entries () >= 2) {
        CT_SimpleName *field = (CT_SimpleName*)((CT_QualName*)node)->Entry (1);
        const char *field_text = field->Text ();
        if (strcmp (field_text, "That") == 0 ||
            strcmp (field_text, "Target") == 0 ||
            strcmp (field_text, "Result") == 0) {
          stringstream field_unit;
          Naming::tjp_typedef (field_unit, field_text);
          const WeavePos &qname_start =
            weave_pos (node->token (), WeavePos::WP_BEFORE);
          const WeavePos &qname_end =
            weave_pos (field->end_token (), WeavePos::WP_AFTER);
          replace (qname_start, qname_end, field_unit.str ());
        }
      }
    }
  }
  else {
    // interate through child nodes
    for (int s = 0; s < node->Sons (); s++)
      provide_typename (node->Son (s));
  }
}


void CodeWeaver::provide_tjp (CFunctionInfo *func, ThisJoinPoint &tjp) {
  assert (func->Tree ());
  assert (func->NextObject ()->FunctionInfo () == func);

  CT_FctDef *fctdef = (CT_FctDef*)func->Tree ();

  // if the function needs the type 'JoinPoint' it has to become a template
  // with 'JoinPoint' as template argument
  if (tjp.type_needed ()) {

    paste (weave_pos (fctdef->token (), WeavePos::WP_BEFORE),
           "template<class JoinPoint> ");

    // also generate some typedefs for That, Target, and Result
    // the corresponding typedef must replace each JoinPoint::XXX
    ostringstream typedefs;
    typedefs << endl;
    typedefs << "  typedef typename JoinPoint::That ";
    Naming::tjp_typedef (typedefs, "That");
    typedefs << ";" << endl;
    typedefs << "  typedef typename JoinPoint::Target ";
    Naming::tjp_typedef (typedefs, "Target");
    typedefs << ";" << endl;
    typedefs << "  typedef typename JoinPoint::Result ";
    Naming::tjp_typedef (typedefs, "Result");
    typedefs << ";" << endl;
    paste (weave_pos (fctdef->Body ()->token (), WeavePos::WP_AFTER),
           typedefs.str ());
  }

  // if a pointer 'tjp' is needed, there has to be an additional argument
  if (tjp.pointer_needed ()) {
    CT_Declarator *last_declarator;
    ((CT_Declarator*)fctdef->Declarator ())->Name (last_declarator);
    CTree *arg_tree = ((CT_FctDeclarator*)last_declarator)->Arguments ();
    ostringstream args;
    args << "JoinPoint *tjp";
    if (func->Arguments () > 0)
      args << ", ";
    paste (weave_pos (arg_tree->token (), WeavePos::WP_AFTER), args.str ());
  }

  // if an alias 'thisJoinPoint' is needed, we must create one
  if (tjp.pointer_alias_needed ()) {
    paste (weave_pos (fctdef->Body ()->token (), WeavePos::WP_AFTER),
           "\n  JoinPoint *&thisJoinPoint = tjp;\n");
  }

  // insert 'typename' in front of JoinPoint::That, Target, ...
  provide_typename (fctdef->Body ());
}


#ifdef ACMODEL
void CodeWeaver::make_advice_call(ostream &out, ACM_Any *loc,
#else
void CodeWeaver::make_advice_call(ostream &out, JoinPointLoc *loc,
#endif
				                          AdviceInfo *ai, bool inter, int depth) {

  CFunctionInfo *srcfunc;
  const char *srcthis_name = "srcthis";

#ifdef ACMODEL
  if (loc->type_val () == JPT_Execution ||
      loc->type_val () == JPT_Construction ||
      loc->type_val () == JPT_Destruction) {
#else
  if (loc->type () == JoinPointLoc::Method ||
      loc->type () == JoinPointLoc::Construction ||
      loc->type () == JoinPointLoc::Destruction) {
#endif
    srcfunc = TransformInfo::of (*loc)->assoc_obj ()->DefObject ()->FunctionInfo ();
    if (!inter)
      srcthis_name = "this";
  }
#ifdef ACMODEL
  else if (loc->type_val () == JPT_Call) {
#else
  else if (loc->type () == JoinPointLoc::MethodCall) {
#endif
    srcfunc = ((TI_MethodCall*)TransformInfo::of (*loc))->caller ();
  }
  else {
    out << "// invalid join point type in make_advice_call" << endl;
    return; // should never happen
  }

  PointCut &pc = ai->pointcut ();
  PointCut::iterator jp_iter = pc.find (loc);
  assert (jp_iter != pc.end ());
  const JoinPoint &jp = *jp_iter;

  if (jp.condition ()) {
    out << "  if(";
    jp.condition ().gen_check (out, ai, srcthis_name);
    out << ") {" << endl;
  }

  // make the advice call itself
  stringstream tjp_tp;
//  make_tjp_typename (tjp_tp, (JPL_Code*)jp.location (), depth);
//  tjp_tp << " ";
  tjp_tp << "__TJP";

  stringstream tjp_obj;
  if (inter)
    tjp_obj << "this";
  else {
    tjp_obj << "&";
    Naming::tjp_instance (tjp_obj, jp.location ());
  }

  out << "  ";
  ai->gen_invocation_func_call (out, tjp_tp.str ().data (),
                                tjp_obj.str ().data ());
  out << endl;

  if (jp.condition ()) {
    out << "  }" << endl;
#ifdef ACMODEL
    if (ai->type () == ACT_AROUND) {
#else
    if (ai->type () == JPL_AdviceCode::ADVICE_AROUND) {
#endif
      out << "  else {" << endl << "    ";
      if (inter)
        out << tjp_tp.str ().data () << "::";
      else {
        Naming::tjp_instance(out, loc);
        out << ".";
      }
      out << "proceed ();" << endl << "  }" << endl;
    }
  }
}


void CodeWeaver::make_advice_calls (ostream &out, JPAdvice *jpa,
#ifdef ACMODEL
  ACM_Any *loc, bool inter) {
#else
  JoinPointLoc *loc, bool inter) {
#endif
  for (int i = 0; i < jpa->before (); i++)
    make_advice_call (out, loc, jpa->before (i), inter, jpa->depth ());

  if (jpa->around ())
    make_advice_call (out, loc, jpa->around (), inter, jpa->depth ());
  else
    make_proceed_code (out, loc, inter, jpa->max_depth ());

  for (int i = 0; i < jpa->after (); i++)
    make_advice_call (out, loc, jpa->after (i), inter, jpa->depth ());
}


// this function weaves the invocation code for a tree of advice functions
// at a call join point
#ifdef ACMODEL
void CodeWeaver::call_join_point (ACM_Call *loc, JPP_Code &plan) {
#else
void CodeWeaver::call_join_point (JPL_MethodCall *loc, JPP_Code &plan) {
#endif
  _code_plan = &plan;

  // determine the weaving parameters
  TI_MethodCall &ti = *(TI_MethodCall*)TransformInfo::of (*loc);
  JPAdvice *cjpa = plan.jp_advice ();
  CObjectInfo *srcobj = TransformInfo::of (*loc)->assoc_obj();
  CFunctionInfo *srcfunc = srcobj->FunctionInfo ();
  CFunctionInfo *dstfunc = ti.called ();
  CTypeInfo *result_type = ti.result_type_info ();

  // analyse the target object of this call
  bool target_is_ptr = false; // initialization to avoid warning
  CT_Expression *target_expr = ti.target_expr (target_is_ptr);

  // determine the weaving position for wrapper code
  const WeavePos *pos = 0;
  bool extern_c_block = false;
  CTree *tree;
  if (srcfunc) {
    CT_FctDef *fctdef = (CT_FctDef*)srcfunc->Tree ();
    tree = linkage_adjust (fctdef);
    if (fctdef->Linkage () && fctdef->Linkage ()->isList ())
      extern_c_block = true;
  }
  else
    tree = ((CT_InitDeclarator*)srcobj->Tree ())->ObjDecl ();
  pos = &weave_pos (tree->token (), WeavePos::WP_BEFORE);

  // generate and insert a forward declaration for the target function if needed
  if (srcfunc && srcfunc == dstfunc && !dstfunc->isMethod () &&
      dstfunc->NextObject () == dstfunc) {
    // TODO: a forward declaration has to be generated and default argument
    //       initializers have to be moved into the forward declaration
//    cout << "FORWARD" << endl;
  }

  // generate JoinPoint-ID-Class and caller function of necessary
  if (ti.needs_rights()) {
    ostringstream bypass;
    const WeavePos *bypass_pos = pos;
    CClassInfo *srccls = srcobj->Scope ()->ClassInfo();
    CClassInfo *dstcls = cscope (dstfunc);

    // check priviledged calls to external functions
    Unit *unit = dstcls->SourceInfo ()->SrcUnit ();
    bool dst_external = false;
    if (!dstcls->SemDB ()->Project ()->isBelow (unit) &&
        !IntroductionUnit::cast (unit)) {
      dst_external = true;
    }

    if (dst_external) {
      // calls to private externals are not supported
      if (dstfunc->Protection() == CProtection::PROT_PRIVATE) {
        _err << sev_error << ti.CallNode ()->token ()->location ()
             << "can't weave call advice for private external function "
             << dstfunc->QualName() << endMessage;
        return;
      }
      // target function is protected
      if (target_expr) {
        CTypeInfo *target_type = target_expr->Type ();
        while (target_type->TypeAddress ())
          target_type = target_type->BaseType ();
        dstcls = target_type->ClassInfo();
      }
      else {
        dstcls = cscope (srcobj);
      }
    }

    bool before_class = (srccls && srcobj->Scope () != dstcls);
    if (before_class) {
      bypass_pos = &weave_pos (srccls->Tree ()->token(), WeavePos::WP_BEFORE);
      close_namespace (bypass, srccls);
    }
    else if (srcobj->Scope()->isNamespace())
      close_namespace (bypass, srcobj);
    string ns_open, ns_close, classname = "Caller";
    if (srcobj->Scope() != dstcls) {
      CObjectInfo *curr = dstcls;
      while (curr->ClassInfo()) {
        classname = string (curr->Name().c_str()) + "::" + classname;
        curr = curr->Scope();
      }
      while (curr && curr->NamespaceInfo() && !curr->FileInfo()) {
        ns_open = string ("namespace ") + curr->Name ().c_str() + " { " + ns_open;
        ns_close += "} ";
        curr = curr->Scope();
      }
    }
    if (ns_open != "")
      bypass << ns_open << endl;
    if (srccls && !before_class)
      bypass << "public:" << endl;
    bypass << "// unique ID-class, just a type" << endl;
    bypass << "struct ";
    Naming::bypass_id_class (bypass, loc);
    bypass << ";" << endl;
    bypass << "template<int X> struct " << classname << "<";
    Naming::bypass_id_class (bypass, loc);
    bypass << ", X> {" << endl;
    bypass << "  static ";
    if (_problems._use_always_inline) // GNU extension
      bypass << "__attribute__((always_inline)) ";
    bypass << "inline ";
    stringstream sig, impl;
    sig << "call (";
    int argnum = 0;
    bool dstthis_needed = needs_this (dstfunc);
    // pass object pointer if needed
    if (dstthis_needed) {
      if (dstfunc->TypeInfo()->isConst())
        sig << "const ";
      stringstream scope;
      dstcls->TypeInfo()->TypeText(scope, "", true, true);
      sig << cleanup_name (scope.str ()) << " *dstthis";
      argnum++;
    }
    // argument list
  #ifdef ACMODEL
    list<ACM_Type*> arg_types;
    get_arg_types (*loc, arg_types);
    int a = 0;
    for (list<ACM_Type*>::iterator i = arg_types.begin(); i != arg_types.end();
         ++i, ++a) {
      CTypeInfo *type = TI_Type::of (**i)->type_info ();
  #else
    for (int a = 0 ; a < loc->arg_count (); a++) {
      CTypeInfo *type = TI_Type::of (loc->arg_type (a))->type_info ();
  #endif
      if (argnum > 0)
        sig << ", ";
      stringstream arg_name;
      arg_name << "arg" << a;
      type->TypeText (sig, arg_name.str ().data (), true, true);
      argnum++;
    }
    sig << ")";
    // add result type
    result_type->TypeText(bypass, sig.str ().data (), true, true);
    bypass << "{" << endl;
    bypass << "    ";
    if (!result_type->isVoid())
      bypass << "return ";
    if (dstthis_needed)
      bypass << "dstthis->";
    // TODO: how to check if a fully qualified name must be used here?
    if (ti.is_qualified())
      bypass << dstfunc->QualName () << " (";
    else
      bypass << dstfunc->Name () << " (";
#ifdef ACMODEL
    a = 0;
    for (list<ACM_Type*>::iterator i = arg_types.begin(); i != arg_types.end();
       ++i, ++a) {
#else
    for (int a = 0 ; a < loc->arg_count (); a++) {
#endif
      if (a > 0)
        bypass << ", ";
      bypass << "arg" << a;
    }
    bypass << ");" << endl;
    bypass << "  }" << endl;
    bypass << "};";
    if (ns_close != "")
      bypass << ns_close << endl;
    if (before_class) {
      open_namespace (bypass, srccls);
    } else if (srcobj->Scope()->isNamespace())
      open_namespace (bypass, srcobj);
    // switch to the right protection if needed
    if (srccls && !before_class && !srccls->isStruct() && !srccls->isUnion() &&
        srcobj->Protection () != CProtection::PROT_NONE &&
        srcobj->Protection () != CProtection::PROT_PUBLIC)
      bypass << protection_string (srcobj->Protection ());

    // finally insert the generated code at the appropriate position
    paste (*bypass_pos, bypass.str ());
  }

  // generate and insert the JoinPoint class
  cjpa->mergeTJPFlags();
  if (cjpa->tjp ().type_needed ()) {
    ostringstream tjp;
    if (extern_c_block)
      tjp << "}" << endl;
    make_tjp_struct (tjp, loc, cjpa);
    // insert the definition
    paste (*pos, tjp.str ());
  }

  // generate replacement function for dstfunc
  ostringstream code;
  code << endl;

  const ThisJoinPoint &tjp = cjpa->tjp();
  bool havesrcthis = srcfunc && needs_this (srcfunc) && (tjp.that() || tjp.useAction());
  bool havedstthis = needs_this (dstfunc) || target_expr;

  // argument list
#ifdef ACMODEL
  int arg_count = get_arg_count (*loc);
#else
  int arg_count = loc->arg_count ();
#endif

  bool wrapper_is_template =
    (ti.has_result () || havesrcthis || havedstthis || arg_count > 0);
  if (wrapper_is_template) {
    code << "template <";
    int targs = 0;
    if (ti.has_result ()) {
      code << "typename TResult";
      targs++;
    }
    if (havesrcthis) {
      if (targs > 0)
        code << ", ";
      code << "typename TSrc";
      targs++;
    }
    if (havedstthis) {
      if (targs > 0)
        code << ", ";
      code << "typename TDst";
      targs++;
    }
    if (arg_count > 0) {
      for (int a = 0; a < arg_count; a++, targs++) {
        if (targs > 0)
          code << ", ";
        code << "typename TArg" << a;
      }
    }
    code << ">" << endl;
  }

  // if the lexical scope in which the associated object of the call was
  // defined is a class scope (not global or namespace) the function definition
  // has to be preceded by a 'static' keyword to avoid an implicit 'this'.
  if (srcobj->Scope ()->ClassInfo ())
      code << "static ";
  if (_problems._use_always_inline) // GNU extension
    code << "__attribute__((always_inline)) ";
  code << "inline ";

  stringstream new_call;
  stringstream asig;
  stringstream impl;

  // add result type
  code << (ti.has_result() ? "TResult " : "void ");

  // add function name
  Naming::call_wrapper (code, loc, cjpa->depth ());
  code << " (";
  Naming::call_wrapper (new_call, loc, cjpa->depth ());
  if (wrapper_is_template) {
    new_call << "< ";
    int targs = 0;
    if (ti.has_result ()) {
      result_type->TypeText(new_call, "", true, true);
      targs++;
    }
    if (havesrcthis) {
      if (targs++ > 0)
        new_call << ", ";
      string scope = string("::") + cscope (srcfunc)->QualName();
      new_call << cleanup_name (scope) << "*";
    }
    if (havedstthis) {
      if (targs++ > 0)
        new_call << ", ";
      if (target_expr) {
        CTypeInfo *target_type = target_expr->Type ();
        while (target_type->TypeAddress ())
          target_type = target_type->BaseType ();
        target_type->TypeText (new_call, target_is_ptr ? "" : "*", true, true);
      }
      else {
        string scope = string ("::") + cscope (srcfunc)->QualName();
        new_call << cleanup_name (scope) << "*";
      }
    }
    if (arg_count > 0) {
      // argument types
#ifdef ACMODEL
      list<ACM_Type*> arg_types;
      get_arg_types (*loc, arg_types);
      int a = 0;
      for (list<ACM_Type*>::iterator i = arg_types.begin(); i != arg_types.end();
           ++i, ++a) {
        CTypeInfo *type = TI_Type::of (**i)->type_info ();
#else
      for (int a = 0 ; a < arg_count; a++) {
        CTypeInfo *type = TI_Type::of (loc->arg_type (a))->type_info ();
#endif
        if (targs++ > 0)
          new_call << ", ";
        type->TypeText (new_call, "", true, true);
      }
    }
    new_call << " >";
  }
  new_call << " (";

  unsigned argnum = 0;
  const char *calltype = ti.CallNode ()->NodeName ();

  // source and destination this pointer arguments (if needed)
  if (havesrcthis) {
    string scope = string("::") + cscope (srcfunc)->QualName();
    code << "TSrc srcthis";
    new_call << "(" << cleanup_name (scope) << "*)this";
    argnum++;
  }

  if (havedstthis) {
    if (argnum > 0)
      code << ", ";
    code << "TDst dstthis";

    // move/paste the target object pointer
    if (argnum > 0)
      new_call << ",";
    if (target_expr) {
      if (!target_is_ptr)
        new_call << "&(";
    }
    else {
      string scope = string ("::") + cscope (srcfunc)->QualName();
      new_call << "(" << cleanup_name (scope) << "*)this";
    }

    argnum++;
  }

  // argument list
  for (int a = 0 ; a < arg_count; a++) {
    if (argnum > 0)
      code << ", ";
    code << "TArg" << a << " arg" << a;
    argnum++;
  }

  code << ")";

  ResultBuffer result_buffer (result_type);

  // replacement function body
  code << "{" << endl;

  // generate __TJP type (for the next level)
  if (cjpa->tjp ().type_needed ()) {
    code << "  typedef ";
    make_tjp_typename (code, loc, cjpa->depth ());
    code << " __TJP;" << endl;
  }

  // declare the result store
  code << "  " << result_buffer.result_declaration();

  // generate common JoinPoint initialization
  make_tjp_common_init(code, loc, cjpa);

  // add calls to advice code
  make_advice_calls(code, cjpa, loc);

  code << "  " << result_buffer.result_return();

  code << " }" << endl;

  if (extern_c_block)
    code << endl << "extern \"C\" {" << endl;

  // now we perform the code manipulation
  // insert the call wrapper body
  paste (*pos, code.str ());

  // check the call for macro problems
  bool macro_problem = false;
  if (calltype == CT_BinaryExpr::NodeId ()) {
  }
  else if (calltype == CT_UnaryExpr::NodeId ()) {
  }
  else if (calltype == CT_PostfixExpr::NodeId ()) {
  }
  else if (calltype == CT_IndexExpr::NodeId ()) {
  }
  else if (calltype == CT_ImplicitCall::NodeId ()) {
  }
  else {
    const char *fctnodetype = ti.CallNode()->Son (0)->NodeName ();
    CT_ExprList *args = ((CT_CallExpr*)ti.CallNode ())->Arguments ();
    if (!args) {
      _err << sev_warning << ti.CallNode ()->token ()->location ()
           << "unsupported kind of call, cannot weave advice code"
           << endMessage;
      return;
    }
    if (fctnodetype == CT_MembRefExpr::NodeId ()) {
      macro_problem =
        kill_macro_problem (((CT_MembRefExpr*)ti.CallNode ()->Son (0))->Son (1)) ||
        kill_macro_problem (((CT_MembRefExpr*)ti.CallNode ()->Son (0))->Son (2));

    }
    else if (fctnodetype == CT_MembPtrExpr::NodeId ()) {
      macro_problem =
        kill_macro_problem (((CT_MembPtrExpr*)ti.CallNode ()->Son (0))->Son (1)) ||
        kill_macro_problem (((CT_MembPtrExpr*)ti.CallNode ()->Son (0))->Son (2));
    }
    else {
      macro_problem =
        kill_macro_problem (ti.CallNode ()->Son (0));
    }
    macro_problem |= kill_macro_problem (args->Son (0));
    macro_problem |= kill_macro_problem (args->Son (args->Sons () - 1));
  }

  // handle problem
  if (macro_problem) {
    if (_problems._warn_macro) {
      _err << sev_warning << ti.CallNode ()->token()->location ()
           << "advice for macro-generated call to '"
           << dstfunc->QualName()
           << "' not woven" << endMessage;
    }
    return;
  }

  // replace call with call to replacement function

  // paste the first (generated) part of the call (in front)
  paste (weave_pos (ti.CallNode ()->token (), WeavePos::WP_BEFORE),
         new_call.str ());

  // paste a closing bracket behind the call
  paste (weave_pos (ti.CallNode ()->end_token (), WeavePos::WP_AFTER),
         ")");

  // provide the other arguments

  CTree *node;
  // call was a binary operator call
  if (calltype == CT_BinaryExpr::NodeId ()) {
    node = ((CT_BinaryExpr*)ti.CallNode ())->Son (0);
    if (target_expr) {
      if (!target_is_ptr)
        paste (weave_pos (node->end_token (), WeavePos::WP_AFTER), ")");
    }
    else if (havesrcthis && !havedstthis)
      paste (weave_pos (node->token (), WeavePos::WP_BEFORE), ",");
    node = ((CT_BinaryExpr*)ti.CallNode ())->Son (1);
    paste (weave_pos (node->end_token (), WeavePos::WP_AFTER), ",");
    kill (node);
  }
  // call was a unary operator call
  else if (calltype == CT_UnaryExpr::NodeId ()) {
    node = ((CT_BinaryExpr*)ti.CallNode ())->Son (0);
    kill (node);
    node = ((CT_BinaryExpr*)ti.CallNode ())->Son (1);
    if (target_expr) {
      if (!target_is_ptr)
        paste (weave_pos (node->end_token (), WeavePos::WP_AFTER), ")");
    }
    else if (havesrcthis && !havedstthis)
      paste (weave_pos (node->token (), WeavePos::WP_BEFORE), ",");
  }
  // call was a postfix operator call
  else if (calltype == CT_PostfixExpr::NodeId ()) {
    node = ((CT_BinaryExpr*)ti.CallNode ())->Son (1);
    kill (node);
    node = ((CT_BinaryExpr*)ti.CallNode ())->Son (0);
    if (strcmp (dstfunc->Name (), "operator ++") == 0 ||
        strcmp (dstfunc->Name (), "operator --") == 0)
      paste (weave_pos (node->end_token (), WeavePos::WP_AFTER), ", 0");
    if (target_expr && !target_is_ptr)
      paste (weave_pos (node->end_token (), WeavePos::WP_AFTER), ")");
  }
  // call was an index operator call
  else if (calltype == CT_IndexExpr::NodeId ()) {
    node = ((CT_BinaryExpr*)ti.CallNode ())->Son (0);
    if (target_expr && !target_is_ptr)
      paste (weave_pos (node->end_token (), WeavePos::WP_AFTER), ")");
    // replace the opening '[' with ','
    node = ((CT_BinaryExpr*)ti.CallNode ())->Son (1);
    paste (weave_pos (node->end_token (), WeavePos::WP_AFTER), ",");
    kill (node);
    // delete the closing ']'
    node = ((CT_BinaryExpr*)ti.CallNode ())->Son (3);
    kill (node);
  }
  // the destination function is a conversion function
  else if (calltype == CT_ImplicitCall::NodeId ()) {
    node = target_expr;
    assert (node);
    paste (weave_pos (node->end_token (), WeavePos::WP_AFTER), ")");
  }
  // call was an ordinary function call
  else {
    const char *fctnodetype = ti.CallNode()->Son (0)->NodeName ();
    CT_ExprList *args = ((CT_CallExpr*)ti.CallNode ())->Arguments ();
    // now perform the transformation
    if ((target_expr && !target_is_ptr) &&
         (fctnodetype == CT_MembRefExpr::NodeId () ||
          fctnodetype == CT_MembPtrExpr::NodeId ())) {
      node = ti.CallNode ()->Son (0);
      paste (weave_pos (node->end_token (), WeavePos::WP_AFTER), ")");
    }
    if (fctnodetype == CT_MembRefExpr::NodeId () && havedstthis) {
      kill (((CT_MembRefExpr*)ti.CallNode ()->Son (0))->Son (1));
      kill (((CT_MembRefExpr*)ti.CallNode ()->Son (0))->Son (2));
    }
    else if (fctnodetype == CT_MembPtrExpr::NodeId () && havedstthis) {
      kill (((CT_MembPtrExpr*)ti.CallNode ()->Son (0))->Son (1));
      kill (((CT_MembPtrExpr*)ti.CallNode ()->Son (0))->Son (2));
    }
    else
      kill (ti.CallNode ()->Son (0));
    if (args->Entries () > 0 && (havedstthis || havesrcthis))
      paste (weave_pos (args->token (), WeavePos::WP_BEFORE), ",");
    kill (args->Son (0));
    kill (args->Son (args->Sons () - 1));
  }
}

void CodeWeaver::insert_bypass_class (CClassInfo *cls) {
  if (!cls->isDefined())
    return;
  CT_ClassDef *cd = (CT_ClassDef*)cls->DefObject ()->Tree ();

  // TODO: ugly workaround needed to check whether a class is defined in
  // an extern "C" block -> extended Puma, SemDB doesn't contain this info!
  for (CTree *curr = cd; curr; curr = curr->Parent())
    if (curr->NodeName() == CT_LinkageSpec::NodeId())
      return;

  Token *open = cd->Members()->token ();
  string bypass_class_str =
      "\n  template <typename, int = 0> struct Caller {};\n"
      "  template <typename, int> friend struct Caller;";
  if (!cls->isStruct() && !cls->isUnion()) { // don't weave this into structs or unions
    bypass_class_str = string ("public:") + bypass_class_str;
    bypass_class_str += string ("\nprivate:");
  }
  paste (weave_pos (open, WeavePos::WP_AFTER), bypass_class_str);
}
